﻿
 //Copyright (c) 2024, Dr Kevin R Bond

Unit PuzzleForm;

Interface

Uses
  qtx.sysutils,
  qtx.classes,
  qtx.dom.types,
  qtx.dom.events,
  qtx.dom.graphics,
  qtx.dom.widgets,
  qtx.dom.border,
  qtx.promises,
  qtx.dom.theme,
  qtx.dom.application,
  qtx.dom.forms,
  qtx.dom.window,
  qtx.dom.events.mouse,
  qtx.dom.events.pointer,
  qtx.dom.events.keyboard,
  qtx.dom.events.touch,
  qtx.delegates,
  qtx.dom.stylesheet,
  qtx.dom.control.label.content,
  qtx.dom.control.contentbox,
  qtx.dom.control.Label,
  qtx.dom.control.panel,
  qtx.dom.control.Button, qtx.dom.http;

Type
  TTileState = (EmptyState, Occupied);

  TEdge = (CTop, CBottom, CLeft, CRight);
  TPuzzleForm = Class;
  TSPuzzle8 = Class;
  TTile = Class (TQTXButton)
            Constructor CreateATile(AOwner : TQTXComponent);
            Procedure Swap(Puzzle8 : TSPuzzle8; Edge : TEdge; RowNo, ColumnNo : Integer);
            Procedure ClearTile(Puzzle8 : TSPuzzle8);
            Function FindWhatState : TTileState;
            Private
              TileState : TTileState;
              y : Integer;
              x : Integer;
              AssignedNo : Integer;
          End;
  TPArrayRow = Array[0..2] Of TTile;
  TPArray = Array[0..2] Of TPArrayRow;
  TEvenPermutationsArray = Array Of String;


  TPuzzleForm = Class(TQTXWindow)
                  {$I "intf::PuzzleForm"}
                  Protected
                    Procedure FormPresented; Override;
                    Procedure FormHidden; Override;
                    Procedure ObjectReady; Override;
                    Procedure StyleObject; Override;
                  Public
                    Constructor Create(AOwner : TQTXComponent; CB : TQTXWindowConstructor); Override;
                    Destructor Destroy; Override;
                  Private
                    Puzzle8 : TSPuzzle8;
                    Procedure CreateEndOfGameWindow(Puzzle8 : TSPuzzle8);
                    Procedure CreateAnotherGoButton(Panel : TQTXPanel; Height, Width : Integer);
                    Procedure CreateQuitButton(Panel : TQTXPanel; Height, Width : Integer);
                End;

 TBlankTilePosition = Record
                        x : Integer;
                        y : Integer;
                      End;

 TSPuzzle8 = Class
               Public
                 Constructor CreateAPuzzle8ObjectLinkedTo(PuzzleForm : TPuzzleForm);
                 Procedure SetUpTilesFor(Puzzle8 : TSPuzzle8);
                 Procedure AssignPermutationTo(Puzzle8 : TSPuzzle8);
                 Procedure LoadJSONIntoPermutationsArrayOf(Puzzle8 : TSPuzzle8);
                 Procedure CheckForPuzzleSolved(Puzzle8 : TSPuzzle8);
                 Procedure MoveTileDownOn(PuzzleForm : TPuzzleForm; NumberTile : TTile; RowNo, ColumnNo : Integer);
                 Procedure MoveTileUpOn(PuzzleForm : TPuzzleForm; NumberTile : TTile; RowNo, ColumnNo : Integer);
                 Procedure MoveTileRightOn(PuzzleForm : TPuzzleForm; NumberTile : TTile; RowNo, ColumnNo : Integer);
                 Procedure MoveTileLeftOn(PuzzleForm : TPuzzleForm; NumberTile : TTile; RowNo, ColumnNo : Integer);
                 Private
                   PuzzleArray : TPArray;
                   EvenPermutationsArray : TEvenPermutationsArray;
                   BlankTilePosition : TBlankTilePosition;
                   Parent : TPuzzleForm;
             End;


Implementation
  Const
    cnNoOfColumns = 3;
    cnNoOfRows = 3;
  Type
    TJEvenPermutation = Class(JObject)
                          Permutation : String;
                        End;
    TJEvenPermutationsArray = Array Of TJEvenPermutation;

  Constructor TTile.CreateATile(AOwner : TQTXComponent);
    Begin
      inherited Create(AOwner, Nil);
    End;

  Function GetState(State : TTileState) : Integer;
    Begin
      Case State Of
        EmptyState : Result := 0;
        Occupied : Result := 1;
      End;
    End;

  Procedure TPuzzleForm.CreateAnotherGoButton(Panel : TQTXPanel; Height, Width : Integer);
    Begin
       TQTXButton.Create(Panel,
                    Procedure (Button : TQTXButton)
                      Begin{****}
                        Button.Width := Width;
                        Button.Height := Height;
                        Button.InnerHtml := '<p style="color: red">Another go?</p>';
                        Button.PositionMode := cpAbsolute;
                        Button.Background.Color := clLightGray;
                        Button.CenterOnParent;
                        Var strPermutation : String = '';
                        Button.OnClick := Procedure (Sender : TObject)
                                            Begin
                                              If Sender Is TQTXButton
                                                Then
                                                  Begin
                                                    Puzzle8.AssignPermutationTo(Self.Puzzle8);
                                                    Panel.Destroy;
                                                  End;
                                            End;
                      End);
    End;

  Procedure TPuzzleForm.CreateQuitButton(Panel : TQTXPanel; Height, Width : Integer);
    Begin
      TQTXButton.Create(Panel,
        Procedure (Button : TQTXButton)
          Begin
            Button.Width := Width;//140;
            Button.Height := Height;//60;
            Button.InnerHtml := '<p style="color: red">Quit?</p>';
            Button.PositionMode := cpAbsolute;
            Button.Background.Color := clLightGray;
            Button.Top :=  TQTXButton(Panel.Items[0]).Top +  TQTXButton(Panel.Items[0]).Height + 20;
            Button.Left :=  TQTXButton(Panel.Items[0]).Left;
            Var strPermutation : String = '';
            Button.OnClick := Procedure (Sender : TObject)
                                Begin
                                  If Sender Is TQTXButton
                                    Then
                                      Begin
                                        Panel.Destroy;
                                        Puzzle8.Parent.Destroy;
                                      End;
                                End;
          End);
    End;

  Procedure TPuzzleForm.CreateEndOfGameWindow(Puzzle8 : TSPuzzle8);
    Begin
      TQTXPanel.Create(Puzzle8.Parent,
        Procedure (Panel : TQTXPanel)
          Begin
            Panel.Width := 300;
            Panel.Height := 300;
            Panel.PositionMode := cpAbsolute;
            Panel.Background.Color := clBlack;
            Panel.CenterOnParent;
            Panel.InnerHTML := '<p style="font-size: 20px; text-align: center; color: white">Puzzle solved, Well done!</p>';
            CreateAnotherGoButton(Panel, 60, 140);
            CreateQuitButton(Panel, 60, 140);
          End);
    End;

  Procedure TSPuzzle8.AssignPermutationTo(Puzzle8 : TSPuzzle8);
    Begin
      Var RandomNo := RandomInt(Puzzle8.EvenPermutationsArray.Length);
      Var strNewPermutation : String := '';
      strNewPermutation := strNewPermutation + Puzzle8.EvenPermutationsArray[RandomNo];
      For Var ColumnNo := 0 To cnNoOfColumns - 1
        Do
          For Var RowNo := 0 To cnNoOfRows - 1
            Do
              Begin
                If Not ((ColumnNo = 2) And (RowNo = 2))
                  Then
                    Begin
                      Puzzle8.PuzzleArray[ColumnNo, RowNo].AssignedNo := StrToInt(strNewPermutation[RowNo*cnNoOfRows + ColumnNo + 1]);
                      Puzzle8.PuzzleArray[ColumnNo, RowNo].InnerText :=  Puzzle8.PuzzleArray[ColumnNo, RowNo].AssignedNo.ToString;
                    End
                  Else Puzzle8.PuzzleArray[2, 2].ClearTile(Puzzle8);
              End;
    End;

  Procedure TSPuzzle8.MoveTileDownOn(PuzzleForm : TPuzzleForm; NumberTile : TTile; RowNo, ColumnNo : Integer);
    Begin
      Var BlankTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo - 1];
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo] := BlankTile;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo - 1] := NumberTile;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo - 1].y := RowNo - 1;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo].y := RowNo;
      NumberTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo - 1];
      NumberTile.Top := 40 + 10*(RowNo - 1) + (RowNo - 1)*NumberTile.Height;
      BlankTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo];
      BlankTile.Top := 40 + 10*RowNo + RowNo*BlankTile.Height;
    End;

  Procedure TSPuzzle8.MoveTileUpOn(PuzzleForm : TPuzzleForm; NumberTile : TTile; RowNo, ColumnNo : Integer);
    Begin
      Var BlankTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo + 1];
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo] := BlankTile;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo + 1] := NumberTile;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo + 1].y := RowNo + 1;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo].y := RowNo;
      NumberTile.Top := 40 + 10*(RowNo + 1) + (RowNo + 1)*NumberTile.Height;
      BlankTile.Top := 40 + 10*RowNo + RowNo*BlankTile.Height;
    End;

  Procedure TSPuzzle8.MoveTileRightOn(PuzzleForm : TPuzzleForm; NumberTile : TTile; RowNo, ColumnNo : Integer);
    Begin
      Var BlankTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo + 1, RowNo];
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo] := BlankTile;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo + 1, RowNo] := NumberTile;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo + 1, RowNo].x := ColumnNo + 1;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo].x := ColumnNo;
      NumberTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo + 1, RowNo];
      NumberTile.Left := 20+ 10*(ColumnNo + 1) + (ColumnNo + 1)*NumberTile.Height;
      BlankTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo];
      BlankTile.Left  := 20+ 10*ColumnNo + ColumnNo*BlankTile.Height;
    End;

  Procedure TSPuzzle8.MoveTileLeftOn(PuzzleForm : TPuzzleForm; NumberTile : TTile; RowNo, ColumnNo : Integer);
    Begin
      Var BlankTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo - 1, RowNo];
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo] := BlankTile;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo - 1, RowNo] := NumberTile;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo - 1, RowNo].x := ColumnNo - 1;
      PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo].x := ColumnNo;
      NumberTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo - 1, RowNo];
      NumberTile.Left := 20+ 10*(ColumnNo - 1) + (ColumnNo - 1)*NumberTile.Height;
      BlankTile := PuzzleForm.Puzzle8.PuzzleArray[ColumnNo, RowNo];
      BlankTile.Left:= 20+ 10*ColumnNo + ColumnNo*BlankTile.Height;
    End;

  Procedure TSPuzzle8.CheckForPuzzleSolved(Puzzle8 : TSPuzzle8);
    Begin
      TQTXDispatch.Execute(Procedure ()
        Begin
          If (Puzzle8.PuzzleArray[0,0].AssignedNo = 1)
              And (Puzzle8.PuzzleArray[1,0].AssignedNo = 2)
                And (Puzzle8.PuzzleArray[2,0].AssignedNo = 3)
                  And (Puzzle8.PuzzleArray[0,1].AssignedNo = 4)
                    And (Puzzle8.PuzzleArray[1,1].AssignedNo = 5)
                      And (Puzzle8.PuzzleArray[2,1].AssignedNo = 6)
                        And (Puzzle8.PuzzleArray[0,2].AssignedNo = 7)
                          And (Puzzle8.PuzzleArray[1,2].AssignedNo = 8)
                            And (Puzzle8.PuzzleArray[2,2].AssignedNo = 0)
            Then Puzzle8.Parent.CreateEndOfGameWindow(Self);
        End, 100);
    End;

  Procedure TTile.Swap(Puzzle8 : TSPuzzle8; Edge : TEdge; RowNo, ColumnNo : Integer);
    Var
      NumberTile, BlankTile : TTile;
      QTXPanel : TQTXPanel;
    Begin
      NumberTile := Puzzle8.PuzzleArray[ColumnNo, RowNo];
      Case Edge Of
        CTop : Puzzle8.MoveTileDownOn(Puzzle8.Parent, NumberTile, RowNo, ColumnNo);
        CBottom : Puzzle8.MoveTileUpOn(Puzzle8.Parent, NumberTile, RowNo, ColumnNo);
        CLeft :  Puzzle8.MoveTileLeftOn(Puzzle8.Parent, NumberTile, RowNo, ColumnNo);
        CRight : Puzzle8.MoveTileRightOn(Puzzle8.Parent, NumberTile, RowNo, ColumnNo);
      End;
      Puzzle8.BlankTilePosition.x := ColumnNo;
      Puzzle8.BlankTilePosition.y := RowNo;
   //   Puzzle8.CheckForPuzzleSolved(Puzzle8);
    End;

  Procedure TTile.ClearTile(Puzzle8 : TSPuzzle8);
    Begin
      Self.InnerText := '';
      Self.Background.Color := clWhite;
      Self.TileState := EmptyState;
      Puzzle8.BlankTilePosition.x := Self.x;
      Puzzle8.BlankTilePosition.y := Self.y;
    End;

  Function TTile.FindWhatState : TTileState;
    Begin
      Result := Occupied;
    End;

  Procedure TSPuzzle8.LoadJSONIntoPermutationsArrayOf(Puzzle8 : TSPuzzle8);
    Var
      Count : Integer;
      strPermutation : String;

    Begin
      TQTXWebLoader.FetchJSON('data/evenpermutations.json',
        Procedure (Uri : string; JSONData : THandle; Error : Exception)
          Begin
            // Deal with an exception
            If Error <> Nil
              Then
                Begin
                  ShowMessage( error.message );
                  Error.Free;
                  Exit;
                End;
            Var j := -1;
            If Not Assigned(TVariant.ToObject(JSONData[0]))
              Then Count := 0
              Else
                Begin
                  Repeat
                    j := j + 1;
                  Until Not Assigned(TVariant.ToObject(JSONData[j]));
                  Count := j;
                  For  Var m := 0 To Count -1
                    Do Puzzle8.EvenPermutationsArray.Add(TJEvenPermutation(TVariant.ToObject(JSONData[m])).Permutation);
                End;
          End);
    End;

  Constructor TSPuzzle8.CreateAPuzzle8ObjectLinkedTo(PuzzleForm : TPuzzleForm);
    Begin
      Parent := PuzzleForm;
      Self.LoadJSONIntoPermutationsArrayOf(Self);
      TQTXDispatch.Execute(Procedure()
        Begin
          SetUpTilesFor(Self);
          AssignPermutationTo(Self);
          Writeln(Self.EvenPermutationsArray.Count);
        End, 1000);
    End;

  Procedure TSPuzzle8.SetUpTilesFor(Puzzle8 : TSPuzzle8);
    Var SwapRowNo, SwapColumnNo : Integer;
    Begin
      For Var RowNo := 0 To cnNoOfRows - 1
        Do
          For Var ColumnNo := 0 To cnNoOfColumns - 1
            Do
              Begin
                Puzzle8.PuzzleArray[ColumnNo, RowNo] := TTile.CreateATile(Puzzle8.Parent);
                Puzzle8.PuzzleArray[ColumnNo, RowNo].PositionMode := cpAbsolute;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].Width := 100;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].Height := 100;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].Left := 18 + 10*ColumnNo + ColumnNo*100;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].Top := 40 + 10*RowNo + RowNo*100;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].x := ColumnNo;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].y := RowNo;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].TileState := Occupied;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].Background.Color := $e3d3d3;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].Border.Color := clGray;
                Puzzle8.PuzzleArray[ColumnNo, RowNo].Font.Size.Value := 20;
                Puzzle8. PuzzleArray[ColumnNo, RowNo].OnClick :=
                  Procedure (Sender : TObject)
                    Var ClickedTile : TTile;
                    Begin
                      If Sender Is TQTXButton
                        Then
                          Begin
                            ClickedTile := Sender As TTile;
                            If ClickedTile.FindWhatState = Occupied
                              Then
                                Begin
                                  SwapRowNo := ClickedTile.y;
                                  SwapColumnNo := ClickedTile.x;
                                  If ((Abs(Puzzle8.BlankTilePosition.x - ClickedTile.x)) = 1)
                                       And ((Abs(Puzzle8.BlankTilePosition.y - ClickedTile.y)) = 0)
                                    Then
                                      If Puzzle8.BlankTilePosition.x < SwapColumnNo
                                        Then ClickedTile.Swap(Puzzle8, CLeft, SwapRowNo, SwapColumnNo)
                                        Else ClickedTile.Swap(Puzzle8, CRight, SwapRowNo, SwapColumnNo)
                                    Else
                                      If (Abs(Puzzle8.BlankTilePosition.y - ClickedTile.y) = 1)
                                            And ((Abs(Puzzle8.BlankTilePosition.x - ClickedTile.x)) = 0)
                                        Then
                                          If Puzzle8.BlankTilePosition.y < SwapRowNo
                                            Then ClickedTile.Swap(Puzzle8, CTop, SwapRowNo, SwapColumnNo)
                                            Else ClickedTile.Swap(Puzzle8, CBottom, SwapRowNo, SwapColumnNo);
                                  Puzzle8.CheckForPuzzleSolved(Puzzle8);
                                End;
                          End;
                     End;
            End;
    End;

  Constructor TPuzzleForm.Create(AOwner : TQTXComponent; CB : TQTXWindowConstructor);
    Begin
      Inherited Create(AOwner, Procedure (PuzzleForm: TQTXWindow)
        Begin
          {$I "impl::PuzzleForm"}
          PuzzleForm.PositionMode := cpAbsolute;
          PuzzleForm.Width := 358;
          PuzzleForm.Height := 400;
          PuzzleForm.CenterOnParent();
          PuzzleForm.Options := [foClose];
          PuzzleForm.Observe := [];
          PuzzleForm.Caption := 'Computable Sliding Puzzle by K R Bond';
          TPuzzleForm(PuzzleForm).Puzzle8 := TSPuzzle8.CreateAPuzzle8ObjectLinkedTo(TPuzzleForm(PuzzleForm));
        End);
    End;

  Destructor TPuzzleForm.Destroy;
    Begin
      Inherited;
    End;

  Procedure TPuzzleForm.StyleObject;
    Begin
      Inherited;
    End;

  Procedure TPuzzleForm.ObjectReady;
    Begin
      Inherited;
    End;

  Procedure TPuzzleForm.FormPresented;
    Begin
      Inherited;
    End;

  Procedure TPuzzleForm.FormHidden;
    Begin
      Inherited;
    End;

Initialization

End.
